import time
import random
import logging
from datetime import datetime, date
import json
from banal.dicts import clean_dict


QUEUE_EXPIRE = 84600 * 14
log = logging.getLogger(__name__)


def backoff(failures=0):
    """Implement a random, growing delay between external service retries."""
    sleep = max(1, failures - 1) + random.random()
    log.debug("Back-off: %.2fs", sleep)
    time.sleep(sleep)


def service_retries():
    """A default number of tries to re-try an external service."""
    return range(30)


def unpack_int(value):
    try:
        return int(value)
    except Exception:
        return 0


def pack_int(value):
    return str(value)


def pack_datetime(value):
    if isinstance(value, (datetime, date)):
        return value.isoformat()


def pack_now():
    return pack_datetime(datetime.utcnow())


def unpack_datetime(value, default=None):
    try:
        return datetime.strptime(value, "%Y-%m-%d %H:%M:%S.%f")
    except Exception:
        try:
            return datetime.strptime(value, "%Y-%m-%d").date()
        except Exception:
            return default


class JSONEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        if isinstance(obj, bytes):
            return obj.decode('utf-8')
        if isinstance(obj, set):
            return [o for o in obj]
        return json.JSONEncoder.default(self, obj)


def dump_json(data):
    if data is None:
        return ''
    data = clean_dict(data)
    return JSONEncoder().encode(data)


def load_json(encoded):
    if encoded is None or encoded == '':
        return
    return json.loads(encoded)



from unittest import TestCase

from servicelayer.cache import get_fakeredis
from servicelayer.jobs import Job, JobStage, RateLimit, Progress, Task


class ProcessTest(TestCase):

    def test_job_queue(self):
        conn = get_fakeredis()
        conn.flushall()
        ds = 'test_1'
        job_id = 'job_1'
        job_stage = JobStage(conn, JobStage.INGEST, job_id, ds)
        status = job_stage.progress.get()
        assert status['pending'] == 0
        assert status['finished'] == 0
        assert job_stage.is_done()
        task1 = Task(job_stage, {'test': 'foo'}, {})
        job_stage.queue_task(task1)
        status = job_stage.progress.get()
        assert status['pending'] == 1
        assert status['finished'] == 0
        assert not job_stage.is_done()
        task = JobStage.get_stage_task(conn, JobStage.INGEST, timeout=None)
        assert task.stage.dataset == job_stage.dataset
        assert task.payload['test'] == 'foo'
        status = job_stage.progress.get()
        assert status['pending'] == 1
        assert status['finished'] == 0
        task2 = Task(job_stage, {'test': 'bar'}, {})
        job_stage.queue_task(task2)
        status = job_stage.progress.get()
        assert status['pending'] == 2
        assert status['finished'] == 0
        task1.done()
        status = job_stage.progress.get()
        assert status['pending'] == 1
        assert status['finished'] == 1
        # Test that pending count remains 1 due to syncing
        task2.done()
        status = job_stage.progress.get()
        assert status['pending'] == 1
        assert status['finished'] == 2
        task = JobStage.get_stage_task(conn, JobStage.INGEST, timeout=1)
        assert task is not None
        assert isinstance(task, Task)
        task = JobStage.get_stage_task(conn, JobStage.INGEST, timeout=1)
        assert task is None

    def test_queue_clear(self):
        conn = get_fakeredis()
        conn.flushall()
        ds = 'test_1'
        job_id = 'job_1'
        job_stage = JobStage(conn, JobStage.INGEST, job_id, ds)
        task = Task(job_stage, {'test': 'foo'}, {})
        job_stage.queue_task(task)
        status = job_stage.progress.get()
        assert status['pending'] == 1
        Job.remove_dataset(conn, ds)
        status = job_stage.progress.get()
        assert status['pending'] == 0
        assert Progress.get_dataset_job_ids(conn, ds) == []

    def test_execute_if_job_done(self):
        conn = get_fakeredis()
        conn.flushall()
        ds = 'test_1'
        job_id = 'job_1'
        job_stage = JobStage(conn, JobStage.INGEST, job_id, ds)
        task1 = Task(job_stage, {'test': 'foo'}, {})
        task2 = Task(job_stage, {'test': 'bar'}, {})
        task1.queue()
        task2.queue()
        status = job_stage.progress.get()
        assert status['pending'] == 2
        test_list = []

        def callback(test_list):
            test_list.append(1)

        task = JobStage.get_stage_task(conn, JobStage.INGEST, timeout=1)
        job_stage.task_done(task)
        status = job_stage.progress.get()
        assert status['pending'] == 1
        job_stage.job.execute_if_done(callback, test_list)
        # callback shouldn't be executed
        print(test_list)
        assert len(test_list) == 0
        task = JobStage.get_stage_task(conn, JobStage.INGEST, timeout=1)
        job_stage.task_done(task)
        status = job_stage.progress.get()
        assert status['pending'] == 0
        job_stage.job.execute_if_done(callback, test_list)
        # callback should have executed
        print(test_list)
        assert len(test_list) == 1

    def test_fetch_multiple_task(self):
        conn = get_fakeredis()
        conn.flushall()
        ds = 'test_1'
        job_id = 'job_1'
        job_stage = JobStage(conn, JobStage.INGEST, job_id, ds)
        task1 = Task(job_stage, {'test': 'foo'}, {})
        task2 = Task(job_stage, {'test': 'bar'}, {})
        job_stage.queue_task(task1)
        job_stage.queue_task(task2)
        status = job_stage.progress.get()
        assert status['pending'] == 2
        tasks = list(job_stage.get_tasks(limit=5))
        assert len(tasks) == 2
        for task in tasks:
            assert isinstance(task, Task)
        assert tasks[0].payload == {'test': 'foo'}
        assert tasks[1].payload == {'test': 'bar'}
        Job.remove_dataset(conn, ds)

    def test_progress(self):
        conn = get_fakeredis()
        ds = 'test_2'
        job_id = 'job_2'
        progress = Progress(conn, JobStage.INGEST, job_id, ds)
        status = progress.get()
        assert status['pending'] == 0
        assert status['finished'] == 0
        progress.mark_pending()
        status = progress.get()
        assert status['pending'] == 1
        assert status['finished'] == 0
        full = Progress.get_dataset_status(conn, ds)
        assert full['pending'] == 1
        job_status = Progress.get_job_status(conn, ds, job_id)
        assert job_status['pending'] == 1
        progress.mark_finished()
        status = progress.get()
        assert status['pending'] == 0
        assert status['finished'] == 1
        job_status = Progress.get_job_status(conn, ds, job_id)
        assert job_status['pending'] == 0
        assert job_status['finished'] == 1
        progress.remove()
        status = progress.get()
        assert status['finished'] == 0

    def test_rate(self):
        conn = get_fakeredis()
        limit = RateLimit(conn, 'banana', limit=10)
        assert limit.check()
        limit.update()
        assert limit.check()
        for i in range(13):
            limit.update()
        assert not limit.check()



import time
import random
import logging
import uuid
from banal import ensure_list

from redis.exceptions import BusyLoadingError, WatchError

from servicelayer.settings import REDIS_LONG
from servicelayer.settings import REDIS_PREFIX as PREFIX
from servicelayer.cache import make_key
from servicelayer.util import pack_now, dump_json, load_json

log = logging.getLogger(__name__)


class Job(object):
    def __init__(self, conn, dataset, job_id):
        self.conn = conn
        self.dataset = dataset
        self.id = job_id
        self.executing_tasks_key = make_key(PREFIX, 'qjt', self.id, dataset)

    @classmethod
    def random_id(cls):
        return uuid.uuid4().hex

    def is_done(self):
        status = Progress.get_job_status(self.conn, self.dataset, self.id)
        return status['pending'] < 1

    def remove(self):
        for stage in Progress.get_dataset_stages(self.conn, self.dataset):  # noqa
            for priority in JobStage.PRIORITIES:
                job_stage = JobStage(
                    self.conn, stage, self.id,
                    self.dataset, priority=priority
                )
                job_stage.remove()

    def _get_all_pending_keys(self):
        """Return the pending keys for all stages in this job"""
        keys = []
        for stage in Progress.get_dataset_stages(self.conn, self.dataset):
            for priority in JobStage.PRIORITIES:
                job_stage = JobStage(
                    self.conn, stage, self.id,
                    self.dataset, priority=priority
                )
                keys.append(job_stage.progress.pending_key)
        return keys

    def execute_if_done(self, callback, *args, **kwargs):
        pipe = self.conn.pipeline()
        try:
            pipe.watch(self.executing_tasks_key, *self._get_all_pending_keys())
            # Make sure there is no tasks currently being executed
            if not self.conn.llen(self.executing_tasks_key) == 0:
                return
            # Make sure none of the stages for the job has non-zero pending
            # tasks
            pending_list = [
                max(0, int(pending or 0)) for pending in pipe.mget(
                    self._get_all_pending_keys()
                )
            ]
            if not all(val == 0 for val in pending_list):
                return
            callback(*args, **kwargs)
        except WatchError:
            log.info("State changed. Not executing callback for job %s", self.id)  # noqa
            return

    def remove_executing_task(self, task):
        self.conn.lrem(self.executing_tasks_key, 1, task.task_id)

    def add_executing_task(self, task):
        self.conn.rpush(self.executing_tasks_key, task.task_id)

    @classmethod
    def remove_dataset(cls, conn, dataset):
        """Delete all known queues associated with a dataset."""
        for job_id in Progress.get_dataset_job_ids(conn, dataset):
            job = cls(conn, dataset, job_id)
            job.remove()


class Task(object):
    def __init__(self, stage, payload=None, context=None, task_id=None):
        self.payload = payload
        self.context = context
        self.stage = stage
        self.job = stage.job
        self.task_id = task_id or self._make_task_id()

    def serialize(self):
        return dump_json({
            'context': self.context or {},
            'payload': self.payload,
            'dataset': self.stage.dataset,
            'job_id': self.job.id,
            'task_id': self.task_id,
            'stage': self.stage.stage,
            'priority': self.stage.priority
        })

    def queue(self):
        self.stage.queue_task(self)

    def done(self):
        self.stage.task_done(self)

    def _make_task_id(self):
        return uuid.uuid4().hex

    @classmethod
    def unpack_taskdata(cls, conn, task_data):
        if task_data is None:
            return None
        task_data = load_json(task_data)
        stage = task_data.get('stage')
        job_id = task_data.get('job_id')
        dataset = task_data.get('dataset')
        priority = task_data.get('priority')
        job_stage = JobStage(conn, stage, job_id, dataset, priority=priority)
        payload = task_data.get('payload')
        context = task_data.get('context')
        task_id = task_data.get('task_id')
        task = Task(job_stage, payload=payload, context=context, task_id=task_id)  # noqa
        # Add task to the list of currently executing tasks
        job = Job(conn, dataset, job_id)
        job.add_executing_task(task)
        return task


class JobStage(object):
    INGEST = 'ingest'

    PRIO_HIGH = 3
    PRIO_MEDIUM = 2
    PRIO_LOW = 1
    PRIORITIES = [PRIO_HIGH, PRIO_MEDIUM, PRIO_LOW]

    def __init__(self, conn, stage, job_id, dataset, priority=PRIO_MEDIUM):  # noqa
        self.conn = conn
        self.stage = stage
        self.job = Job(conn, dataset, job_id)
        self.dataset = dataset
        self.priority = priority
        self.progress = Progress(conn, stage, self.job.id, dataset)
        self.queue_key = make_key(PREFIX, 'q', stage, self.job.id, dataset)
        self.ops_queue_key = make_key(PREFIX, 'qo', stage, priority)

    def queue_task(self, task):
        self.conn.sadd(self.ops_queue_key, self.queue_key)
        data = task.serialize()
        self.conn.rpush(self.queue_key, data)
        self.progress.mark_pending()

    def task_done(self, task):
        self.job.remove_executing_task(task)
        self.progress.mark_finished()
        self.sync()

    def _sync(self, pipe):
        pending = pipe.llen(self.queue_key)
        pipe.set(self.progress.pending_key, pending)

    def sync(self):
        self.conn.transaction(self._sync, self.queue_key)

    def is_done(self):
        """Are the tasks for the current `job_id` and `stage` done?"""
        status = self.progress.get()
        return status.get('pending') < 1

    def remove(self):
        """Remove tasks for the current `job_id` and `stage`"""
        self.conn.srem(self.ops_queue_key, self.queue_key)
        self.conn.delete(self.queue_key)
        self.progress.remove()

    def is_job_done(self):
        """Are all the tasks for `job_id` done?"""
        return self.job.is_done()

    def remove_job(self):
        """Remove all tasks for `job_id`"""
        return self.job.remove()

    def get_tasks(self, limit=500):
        assert limit > 0
        pipe = self.conn.pipeline()
        pipe.lrange(self.queue_key, 0, limit-1)
        pipe.ltrim(self.queue_key, limit, -1)
        tasks = pipe.execute()[0]
        for task in tasks:
            yield Task.unpack_taskdata(self.conn, task)

    @classmethod
    def _get_stage_queues(cls, conn, stages):
        """Return all the active queues for the given stage."""
        queues = []
        for priority in cls.PRIORITIES:
            prio_queues = []
            for op in ensure_list(stages):
                key = make_key(PREFIX, 'qo', op, priority)
                prio_queues.extend(conn.smembers(key))
            # TODO: do we want to random.shuffle?
            random.shuffle(prio_queues)
            for queue in prio_queues:
                if queue not in queues:
                    queues.append(queue)
        return queues

    @classmethod
    def get_stage_task(cls, conn, stages, timeout=0):
        """Retrieve a single task from the highest-priority queue that has
        work pending."""
        try:
            queues = cls._get_stage_queues(conn, stages)
            if not len(queues):
                return None
            # Support a magic value to not block, i.e. timeout None
            if timeout is None:
                # LPOP does not support multiple lists.
                for queue in queues:
                    task_data = conn.lpop(queue)
                    if task_data is not None:
                        break
            else:
                task_data = conn.blpop(queues, timeout=timeout)
                if task_data is not None:
                    key, task_data = task_data

            return Task.unpack_taskdata(conn, task_data)
        except BusyLoadingError:
            time.sleep(timeout + 1)
            return None


class Progress(object):
    """An abstraction for the notion of a process that covers a set
    of items. This is simplified and does not cover the notion of a
    task being in-progress."""
    PENDING = 'pending'
    FINISHED = 'finished'
    # TODO: do we need a 'running' state?

    def __init__(self, conn, stage, job_id, dataset):
        _key = make_key(PREFIX, 'p', stage, job_id, dataset)
        self.conn = conn
        self.stage = stage
        self.job_id = job_id
        self.dataset = dataset
        self.dataset_key = make_key(PREFIX, 'qd', dataset)
        self.dataset_jobs_key = make_key(PREFIX, 'qdj', dataset)
        self.pending_key = make_key(_key, self.PENDING)
        self.finished_key = make_key(_key, self.FINISHED)

    def mark_active(self):
        """Add the dataset to the list of datasets that are running. Add the
        job and it's start time if it doesn't exist already."""
        self.conn.sadd(self.dataset_key, self.stage)
        if not self.conn.hget(self.dataset_jobs_key, self.job_id):
            self.conn.hset(self.dataset_jobs_key, self.job_id, pack_now())

    def mark_pending(self, amount=1):
        self.mark_active()
        self.conn.incr(self.pending_key, amount=amount)

    def mark_finished(self, amount=1):
        """Move items from the pending to the finished set."""
        pipe = self.conn.pipeline()
        pipe.decr(self.pending_key, amount=amount)
        pipe.incr(self.finished_key, amount=amount)
        pipe.execute()

    def put_finished(self, amount=1):
        """Add a number of items to the finished set without reducing
        the count of pending items."""
        self.mark_active()
        self.conn.incr(self.finished_key, amount=amount)

    def remove(self):
        pipe = self.conn.pipeline()
        pipe.srem(self.dataset_key, self.stage)
        pipe.expire(self.dataset_key, REDIS_LONG)
        pipe.hdel(self.dataset_jobs_key, self.job_id)
        pipe.delete(self.pending_key, self.finished_key)
        pipe.execute()

    def get(self):
        """Get the current status."""
        keys = (self.pending_key, self.finished_key)
        pending, finished = self.conn.mget(keys)
        return {
            'job_id': self.job_id,
            'stage': self.stage,
            'pending': max(0, int(pending or 0)),
            'finished': max(0, int(finished or 0)),
        }

    @classmethod
    def get_dataset_stages(cls, conn, dataset):
        return conn.smembers(make_key(PREFIX, 'qd', dataset))

    @classmethod
    def get_dataset_job_ids(cls, conn, dataset):
        return conn.hkeys(make_key(PREFIX, 'qdj', dataset))

    @classmethod
    def get_job_status(cls, conn, dataset, job_id):
        """Aggregate status for all stages on the given job."""
        status = {'finished': 0, 'pending': 0, 'stages': []}
        status['start_time'] = conn.hget(make_key(PREFIX, 'qdj', dataset), job_id)  # noqa
        for stage in cls.get_dataset_stages(conn, dataset):
            progress = cls(conn, stage, job_id, dataset).get()
            status['stages'].append(progress)
            status['finished'] += progress['finished']
            status['pending'] += progress['pending']
        return status

    @classmethod
    def get_dataset_status(cls, conn, dataset):
        """Aggregate status for all stages on the given dataset."""
        status = {'finished': 0, 'pending': 0, 'jobs': []}
        for job_id in cls.get_dataset_job_ids(conn, dataset):
            progress = cls.get_job_status(conn, dataset, job_id)
            status['jobs'].append(progress)
            status['finished'] += progress['finished']
            status['pending'] += progress['pending']
        return status


class RateLimit(object):
    """Limit the rate of stages on a given resource during a
    stated interval."""

    def __init__(self, conn, resource, limit=100, interval=60, unit=1):
        self.conn = conn
        self.resource = resource
        self.limit = max(0.1, limit)
        self.interval = max(1, interval)
        self.unit = unit

    def _time(self):
        return int(time.time() / self.unit)

    def _keys(self):
        base = self._time()
        for slot in range(base, base + self.interval):
            yield make_key(PREFIX, 'rate', self.resource, slot)

    def update(self, amount=1):
        """Set the cached counts for stats keeping."""
        pipe = self.conn.pipeline()
        for key in self._keys():
            pipe.incr(key, amount=amount)
            pipe.expire(key, (self.interval * self.unit) + 2)
        values = pipe.execute()[::2]
        return (sum(values) / self.interval)

    def check(self):
        """Check if the resource has exceeded the rate limit."""
        key = make_key(PREFIX, 'rate', self.resource, self._time())
        count = int(self.conn.get(key) or 0)
        return count < self.limit

    def comply(self, amount=1):
        """Update, then sleep for the time required to adhere to the
        rate limit."""
        # FIXME: this is trying to be smart, which will probably
        # backfire. Rather than just halt when the rate is exceeded,
        # this is trying to average down calls to the appropriate
        # frequency.
        rate = self.update(amount=amount) / self.limit
        expected = self.interval / self.limit
        excess = rate - expected
        if excess > 0:
            time.sleep(excess)
        return rate



import signal
import logging
from threading import Thread
from abc import ABC, abstractmethod

from servicelayer import settings
from servicelayer.jobs import JobStage
from servicelayer.cache import get_redis

log = logging.getLogger(__name__)


class Worker(ABC):
    """Workers of all microservices, unite!"""

    def __init__(self, conn=None, stages=None,
                 num_threads=settings.WORKER_THREADS):
        self.conn = conn or get_redis()
        self.stages = stages
        self.num_threads = num_threads
        self._shutdown = False

    def shutdown(self, *args):
        log.warning("Shutting down worker.")
        self._shutdown = True
        if not self.num_threads:
            raise SystemExit()

    def handle_safe(self, task):
        try:
            self.handle(task)
        except (SystemExit, KeyboardInterrupt, Exception):
            self.retry(task)
            raise
        finally:
            task.done()
            if task.job.is_done():
                task.stage.sync()
            self.after_task(task)

    def init_internal(self):
        self._shutdown = False
        self.boot()

    def retry(self, task):
        retries = int(task.context.get('retries', 0))
        if retries < settings.WORKER_RETRY:
            log.warning("Queue failed task for re-try...")
            task.context['retries'] = retries + 1
            task.stage.queue_task(task)

    def process(self):
        while True:
            self.periodic()
            task = JobStage.get_stage_task(self.conn,
                                           self.get_stages(),
                                           timeout=5)
            if task is None:
                continue
            if self._shutdown:
                self.retry(task)
                return
            self.handle_safe(task)
            if self._shutdown:
                return

    def sync(self):
        """Process only the tasks already in the job queue, but do not
        go into an infinte loop waiting for new ones."""
        self.init_internal()
        while True:
            task = JobStage.get_stage_task(self.conn,
                                           self.get_stages(),
                                           timeout=1)
            if task is None:
                return
            self.handle_safe(task)

    def run(self):
        # Try to quit gracefully, e.g. after finishing the current task.
        signal.signal(signal.SIGINT, self.shutdown)
        signal.signal(signal.SIGTERM, self.shutdown)
        self.init_internal()
        if not self.num_threads:
            return self.process()
        log.info("Worker has %d threads.", self.num_threads)
        threads = []
        for _ in range(self.num_threads):
            thread = Thread(target=self.process)
            thread.daemon = True
            thread.start()
            threads.append(thread)
        for thread in threads:
            thread.join()

    def get_stages(self):
        """Easily allow the user to make the active stages dynamic."""
        return self.stages

    def boot(self):
        """Optional hook for the boot-up of the worker."""
        pass

    def periodic(self):
        """Optional hook for running a periodic task checker."""
        pass

    def after_task(self, task):
        """Optional hook excuted after handling a task"""
        pass

    @abstractmethod
    def handle(self, task):
        raise NotImplementedError



from unittest import TestCase
from datetime import datetime

from servicelayer.util import pack_datetime, pack_now, unpack_datetime
from servicelayer.util import pack_int, unpack_int
from servicelayer.util import dump_json, load_json


class UtilTest(TestCase):

    def test_int(self):
        packed = pack_int(555)
        unpacked = unpack_int(packed)
        assert unpacked == 555
        assert unpack_int('Banana') == 0

    def test_datetime(self):
        date = datetime.utcnow().date()
        packed = pack_datetime(date)
        unpacked = unpack_datetime(packed)
        assert unpacked == date
        now = pack_now()
        assert now.startswith(packed), now
        assert pack_datetime('Banana') is None
        assert unpack_datetime('Banana') is None

    def test_object(self):
        data = {
            'help': 5,
            'xxx': datetime.utcnow(),
            'bla': None,
            'foo': set(['a']),
            'yum': b'xxx'
        }
        json = dump_json(data)
        out = load_json(json)
        assert 'bla' not in out, out
        assert 'xxx' in out, out
        assert 'a' in out['foo'], out
        assert load_json('') is None
        assert load_json(None) is None
        assert dump_json(None) == ''



from unittest import TestCase

from servicelayer.cache import get_fakeredis
from servicelayer.jobs import JobStage, Task
from servicelayer.worker import Worker


class CountingWorker(Worker):

    def boot(self):
        self.test_done = 0

    def handle(self, task):
        self.test_done += 1


class WorkerTest(TestCase):

    def test_run(self):
        conn = get_fakeredis()
        operation = 'lala'
        worker = CountingWorker(conn=conn, stages=[operation])
        worker.sync()
        assert worker.test_done == 0, worker.test_done
        stage = JobStage(conn, operation, 'job_id', 'test')
        task = Task(stage, {}, {})
        task.queue()
        assert not stage.is_done()
        assert worker.test_done == 0, worker.test_done
        worker.sync()
        assert worker.test_done == 1, worker.test_done
        assert stage.is_done()
        worker.retry(task)
        assert not stage.is_done()
        worker.sync()
        assert stage.is_done()
        assert worker.test_done == 1, worker.test_done
        worker.shutdown()
        worker.retry(task)
        worker.process()



import shutil
import logging
from normality import safe_filename

from servicelayer.archive.archive import Archive
from servicelayer.archive.util import ensure_path, checksum, BUF_SIZE

log = logging.getLogger(__name__)


class FileArchive(Archive):

    def __init__(self, path=None):
        self.path = ensure_path(path)
        if self.path is None:
            raise ValueError('No archive path is set.')
        log.info("Archive: %s", self.path)

    def _locate_key(self, content_hash):
        prefix = self._get_prefix(content_hash)
        if prefix is None:
            return
        path = self.path.joinpath(prefix)
        try:
            for file_name in path.iterdir():
                return file_name.resolve()
        except FileNotFoundError:
            return

    def archive_file(self, file_path, content_hash=None, mime_type=None):
        """Import the given file into the archive."""
        if content_hash is None:
            content_hash = checksum(file_path)

        if content_hash is None:
            return

        if self._locate_key(content_hash):
            return content_hash

        archive_prefix = self._get_prefix(content_hash)
        archive_path = self.path.joinpath(archive_prefix)
        archive_path.mkdir(parents=True, exist_ok=True)
        file_name = safe_filename(file_path, default='data')
        archive_path = archive_path.joinpath(file_name)
        with open(file_path, 'rb') as fin:
            with open(archive_path, 'wb') as fout:
                shutil.copyfileobj(fin, fout, BUF_SIZE)
        return content_hash

    def load_file(self, content_hash, file_name=None, temp_path=None):
        return self._locate_key(content_hash)
